from pathlib import Path

import flask
from packaging.utils import canonicalize_name

BASE_DIR = Path(__file__).parent
bp = flask.Blueprint("main", __name__)

INDEX_TEMPLATE = """\
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="utf-8">
    <title>Test PyPI Simple Index</title>
</head>
<body>
    <h1>Test PyPI Simple Index</h1>
    {% for package in packages-%}
    <a href="{{ url_for('.package_index', package=package) }}">{{ package }}</a><br/>
    {%-endfor %}
</body>
</html>
"""


@bp.route("/files/<path:path>")
def package_file(path):
    return flask.send_from_directory(BASE_DIR / "files", path)


@bp.route("/simple/<package>")
def package_index(package):
    canonical_name = canonicalize_name(package)
    if package != canonical_name:
        return flask.redirect(flask.url_for(".package_index", package=canonical_name))
    return flask.send_from_directory(BASE_DIR / "index", package + ".html")


@bp.route("/simple")
def package_index_root():
    packages = [p.stem for p in (BASE_DIR / "index").glob("*.html")]
    return flask.render_template_string(INDEX_TEMPLATE, packages=packages)


@bp.route("/links/index.html")
def find_links():
    return flask.send_from_directory(BASE_DIR / "findlinks", "index.html")


def create_app():
    app = flask.Flask(__name__)
    app.url_map.strict_slashes = False
    app.register_blueprint(bp)
    return app


if __name__ == "__main__":
    app = create_app()
    app.run(port=8000)



from __future__ import annotations

from unearth.link import Link


class URLError(ValueError):
    pass


class VCSBackendError(URLError):
    pass


class UnpackError(RuntimeError):
    pass


class HashMismatchError(UnpackError):
    def __init__(
        self, link: Link, expected: dict[str, list[str]], actual: dict[str, str]
    ) -> None:
        self.link = link
        self.expected = expected
        self.actual = actual

    def format_hash_item(self, name: str) -> str:
        expected = self.expected[name]
        actual = self.actual[name]
        expected_prefix = f"Expected({name}): "
        actual_prefix = f"  Actual({name}): "
        sep = "\n" + " " * len(expected_prefix)
        return f"{expected_prefix}{sep.join(expected)}\n{actual_prefix}{actual}"

    def __str__(self) -> str:
        return f"Hash mismatch for {self.link.redacted}:\n" + "\n".join(
            self.format_hash_item(name) for name in sorted(self.expected)
        )



from pathlib import Path

import pytest

from unearth.link import Link
from unearth.utils import WINDOWS


@pytest.mark.parametrize(
    "url, normalized",
    [
        ("https://pypi.org/simple", "https://pypi.org/simple"),
        (
            "git+https://github.com/pallets/click.git@master",
            "git+https://github.com/pallets/click.git@master",
        ),
        ("git+git@github.com:pypa/pip.git", "git+ssh://git@github.com/pypa/pip.git"),
    ],
)
def test_link_normalized(url, normalized):
    assert Link(url).normalized == normalized


@pytest.mark.parametrize(
    "left,right,is_equal",
    [
        (
            Link("https://pypi.org/simple"),
            Link("https://pypi.org/simple", comes_from="https://pypi.org"),
            True,
        ),
        (
            Link("git+git@github.com:pypa/pip.git"),
            Link("git+ssh://git@github.com/pypa/pip.git"),
            True,
        ),
        (
            Link("git+https://github.com/pypa/pip.git@main"),
            Link("git+https://github.com/pypa/pip.git@22.0.4"),
            False,
        ),
        (
            Link("git+https://github.com/pypa/pip.git@main#egg=pip"),
            Link("git+https://github.com/pypa/pip.git@main#egg=pip&subdirectory=src"),
            False,
        ),
        (
            Link("https://pypi.org/simple/click/8.0.1"),
            Link("https://pypi.org/simple/click/8.0.1", yank_reason=""),
            False,
        ),
        (
            Link("https://pypi.org/simple/click/8.0.1", requires_python=">=3.6"),
            Link("https://pypi.org/simple/click/8.0.1", requires_python=">=3.7"),
            False,
        ),
    ],
)
def test_link_equality(left, right, is_equal):
    assert (left == right) is is_equal


def test_link_is_file_and_filepath():
    if WINDOWS:
        link = Link("file:///C:/path/to/file")
        path = Path("C:/path/to/file")
    else:
        link = Link("file:///path/to/file")
        path = Path("/path/to/file")
    assert link.is_file
    assert link.file_path == path

    assert not Link("http://example.org/").is_file


@pytest.mark.parametrize(
    "url,expected",
    [
        ("http://example.org/", False),
        ("git+git@github.com:pypa/pip.git", True),
        ("svn+https://svn.example.org/repo", True),
        ("abc+https://test.com/", False),
    ],
)
def test_link_is_vcs(url, expected):
    assert Link(url).is_vcs == expected


def test_link_url_without_fragment():
    link = Link("https://pypi.org/simple#egg=pip&subdirectory=src")
    assert link._fragment_dict == {"egg": "pip", "subdirectory": "src"}
    assert link.subdirectory == "src"
    assert link.url_without_fragment == "https://pypi.org/simple"


def test_link_filename_and_hash():
    link = Link(
        "https://files.pythonhosted.org/packages/c2/f1/df59e28c642d583f7dacffb1e0965d"
        "0e00b218e0186d7858ac5233dce840/click-8.1.3%2Bcpu-py3-none-any.whl"
        "#sha256=bb4d8133cb15a609f44e8213d9b391b0809795062913b383c62be0ee95b1db48"
    )
    assert link.filename == "click-8.1.3+cpu-py3-none-any.whl"
    assert link.is_wheel
    assert link.hash_name == "sha256"
    assert (
        link.hash == "bb4d8133cb15a609f44e8213d9b391b0809795062913b383c62be0ee95b1db48"
    )


@pytest.mark.parametrize(
    "url,splitted,redacted",
    (
        [
            (
                "https://pypi.org/simple",
                (None, "https://pypi.org/simple"),
                "https://pypi.org/simple",
            ),
            (
                "https://abc@pypi.org/simple",
                (("abc", None), "https://pypi.org/simple"),
                "https://***@pypi.org/simple",
            ),
            (
                "https://abc:pass@pypi.org/simple",
                (("abc", "pass"), "https://pypi.org/simple"),
                "https://***@pypi.org/simple",
            ),
        ]
    ),
)
def test_link_split_auth_and_redact(url, splitted, redacted):
    link = Link(url)
    assert splitted == link.split_auth()
    assert link.redacted == redacted



import pytest

from unearth import Link
from unearth.evaluator import TargetPython
from unearth.finder import PackageFinder

pytestmark = pytest.mark.usefixtures("pypi")


@pytest.mark.parametrize(
    "target_python,filename",
    [
        (
            TargetPython((3, 9), abis=["cp39"], impl="cp", platforms=["win_amd64"]),
            "black-22.3.0-cp39-cp39-win_amd64.whl",
        ),
        (
            TargetPython((3, 8), abis=["cp38"], impl="cp", platforms=["win_amd64"]),
            "black-22.3.0-cp38-cp38-win_amd64.whl",
        ),
        (
            TargetPython(
                (3, 9), abis=["cp39"], impl="cp", platforms=["manylinux2014_x86_64"]
            ),
            "black-22.3.0-cp39-cp39-manylinux_2_17_x86_64.manylinux2014_x86_64.whl",
        ),
        (
            TargetPython(
                (3, 9), abis=["cp39"], impl="cp", platforms=["macosx_11_0_arm64"]
            ),
            "black-22.3.0-py3-none-any.whl",
        ),
    ],
)
def test_find_most_matching_wheel(session, target_python, filename):
    finder = PackageFinder(
        session, index_urls=["https://pypi.org/simple"], target_python=target_python
    )
    assert finder.find_best_match("black").best.link.filename == filename


def test_find_package_with_format_control(session):
    finder = PackageFinder(
        session,
        index_urls=["https://pypi.org/simple"],
        target_python=TargetPython(
            (3, 9), abis=["cp39"], impl="cp", platforms=["win_amd64"]
        ),
        no_binary=["black"],
        only_binary=["first"],
    )
    assert finder.find_best_match("black").best.link.filename == "black-22.3.0.tar.gz"
    assert (
        finder.find_best_match("first").best.link.filename
        == "first-2.0.1-py2.py3-none-any.whl"
    )


def test_find_package_no_binary_for_all(session):
    finder = PackageFinder(
        session,
        index_urls=["https://pypi.org/simple"],
        target_python=TargetPython(
            (3, 9), abis=["cp39"], impl="cp", platforms=["win_amd64"]
        ),
        no_binary=[":all:"],
    )
    assert finder.find_best_match("black").best.link.filename == "black-22.3.0.tar.gz"
    assert finder.find_best_match("first").best.link.filename == "first-2.0.2.tar.gz"


def test_find_package_prefer_binary(session):
    finder = PackageFinder(
        session,
        index_urls=["https://pypi.org/simple"],
        target_python=TargetPython(
            (3, 9), abis=["cp39"], impl="cp", platforms=["win_amd64"]
        ),
        prefer_binary=True,
    )
    assert (
        finder.find_best_match("first").best.link.filename
        == "first-2.0.1-py2.py3-none-any.whl"
    )


def test_find_package_with_hash_allowance(session):
    finder = PackageFinder(
        session,
        index_urls=["https://pypi.org/simple"],
        target_python=TargetPython(
            (3, 9), abis=["cp39"], impl="cp", platforms=["win_amd64"]
        ),
    )
    assert (
        finder.find_best_match(
            "first",
            hashes={
                "sha256": [
                    "41d5b64e70507d0c3ca742d68010a76060eea8a3d863e9b5130ab11a4a91aa0e"
                ]
            },
        ).best.link.filename
        == "first-2.0.1-py2.py3-none-any.whl"
    )


@pytest.mark.parametrize("ignore_compat", [True, False])
def test_find_package_ignoring_compatibility(session, ignore_compat):
    finder = PackageFinder(
        session,
        index_urls=["https://pypi.org/simple"],
        target_python=TargetPython(
            (3, 9), abis=["cp39"], impl="cp", platforms=["win_amd64"]
        ),
        ignore_compatibility=ignore_compat,
    )
    all_available = finder.find_matches("black==22.3.0")
    assert len(all_available) == (6 if ignore_compat else 3)


def test_find_package_with_version_specifier(session):
    finder = PackageFinder(
        session,
        index_urls=["https://pypi.org/simple"],
        ignore_compatibility=True,
    )
    matches = finder.find_matches("black==22.3.0")
    assert len(matches) == 6
    assert all(p.name == "black" and p.version == "22.3.0" for p in matches)
    matches = finder.find_matches("black<22.3.0")
    assert len(matches) == 0


def test_find_package_allowing_prereleases(session):
    finder = PackageFinder(
        session,
        index_urls=["https://pypi.org/simple"],
        ignore_compatibility=True,
    )
    matches = finder.find_matches("black<22.3.0", allow_prereleases=True)
    assert len(matches) == 2
    assert all(p.name == "black" and p.version == "21.12b0" for p in matches)

    matches = finder.find_matches("black==21.12b0")
    assert len(matches) == 2
    assert all(p.name == "black" and p.version == "21.12b0" for p in matches)

    matches = finder.find_matches("black<=21.12b1", allow_prereleases=False)
    assert len(matches) == 0


def test_find_requirement_with_link(session):
    finder = PackageFinder(
        session,
        index_urls=["https://pypi.org/simple"],
        ignore_compatibility=True,
    )
    req = "first @ https://pypi.org/files/first-2.0.2.tar.gz"
    matches = finder.find_matches(req)
    assert len(matches) == 1
    assert matches[0].name == "first"
    assert matches[0].link.normalized == "https://pypi.org/files/first-2.0.2.tar.gz"


def test_find_requirement_preference(session, fixtures_dir):
    find_link = Link.from_path(fixtures_dir / "findlinks/index.html")
    finder = PackageFinder(
        session,
        index_urls=["https://pypi.org/simple"],
        find_links=[find_link.normalized],
        ignore_compatibility=True,
    )
    best = finder.find_best_match("first").best
    assert best.link.filename == "first-2.0.3-py2.py3-none-any.whl"
    assert best.link.comes_from == find_link.normalized


def test_find_requirement_preference_respect_source_order(session, fixtures_dir):
    find_link = Link.from_path(fixtures_dir / "findlinks/index.html")
    finder = PackageFinder(
        session,
        index_urls=["https://pypi.org/simple"],
        find_links=[find_link.normalized],
        ignore_compatibility=True,
        respect_source_order=True,
    )
    best = finder.find_best_match("first").best
    assert best.link.filename == "first-2.0.2.tar.gz"
    assert best.link.comes_from == "https://pypi.org/simple/first/"



"""The main class for the unearth package."""
from __future__ import annotations

import atexit
import functools
import os
import pathlib
from tempfile import TemporaryDirectory
from typing import Iterable, NamedTuple, cast
from urllib.parse import urljoin

import packaging.requirements
from packaging.utils import BuildTag, canonicalize_name, parse_wheel_filename
from packaging.version import parse as parse_version

from unearth.collector import collect_links_from_location
from unearth.evaluator import (
    Evaluator,
    FormatControl,
    Package,
    TargetPython,
    evaluate_package,
    is_equality_specifier,
)
from unearth.link import Link
from unearth.preparer import unpack_link
from unearth.session import PyPISession
from unearth.utils import split_auth_from_url


class BestMatch(NamedTuple):
    """The best match for a package."""

    #: The best matching package, or None if no match was found.
    best: Package | None
    #: The applicable packages, excluding those with unmatching versions.
    applicable: list[Package]
    #: All candidates found for the requirement.
    candidates: list[Package]


class PackageFinder:
    """The main class for the unearth package.

    Args:
        session (PyPISession|None): The session to use for the finder.
            If not provided, a temporary session will be created.
        index_urls: (Iterable[str]): The urls of the index pages.
        find_links: (Iterable[str]): The urls or paths of the find links.
        trusted_hosts: (Iterable[str]): The trusted hosts.
        target_python (TargetPython): The links must match
            the target Python
        ignore_compatibility (bool): Whether to ignore the compatibility check
        no_binary (Iterable[str]): The names of the packages to disallow wheels
        only_binary (Iterable[str]): The names of the packages to disallow non-wheels
        prefer_binary (bool): Whether to prefer binary packages even if
            newer sdist pacakges exist.
        respect_source_order (bool): If True, packages from the source coming earlier
            are more preferred, even if they have lower versions.
        verbosity (int): The verbosity level.
    """

    def __init__(
        self,
        session: PyPISession | None = None,
        index_urls: Iterable[str] = (),
        find_links: Iterable[str] = (),
        trusted_hosts: Iterable[str] = (),
        target_python: TargetPython | None = None,
        ignore_compatibility: bool = False,
        no_binary: Iterable[str] = (),
        only_binary: Iterable[str] = (),
        prefer_binary: bool = False,
        respect_source_order: bool = False,
        verbosity: int = 0,
    ) -> None:
        self.index_urls = list(index_urls)
        self.find_links = list(find_links)
        self.target_python = target_python or TargetPython()
        self.ignore_compatibility = ignore_compatibility
        self.no_binary = [canonicalize_name(name) for name in no_binary]
        self.only_binary = [canonicalize_name(name) for name in only_binary]
        self.prefer_binary = prefer_binary
        if session is None:
            session = PyPISession(
                index_urls=self.index_urls, trusted_hosts=trusted_hosts
            )
            atexit.register(session.close)
        self.session = session
        self.respect_source_order = respect_source_order
        self.verbosity = verbosity

        self._tag_priorities = {
            tag: i for i, tag in enumerate(self.target_python.supported_tags())
        }
        # Index pages are preferred over find links.
        self._source_order = [
            split_auth_from_url(url)[1] for url in (self.index_urls + self.find_links)
        ]

    def build_evaluator(
        self,
        package_name: str,
        allow_yanked: bool = False,
        hashes: dict[str, list[str]] | None = None,
    ) -> Evaluator:
        """Build an evaluator for the given package name.

        Args:
            package_name (str): The desired package name
            allow_yanked (bool): Whether to allow yanked candidates.
            hashes (dict[str, list[str]]|None): The hashes to filter on.

        Returns:
            Evaluator: The evaluator for the given package name
        """
        if hashes:
            hashes = {name: sorted(values) for name, values in hashes.items()}
        canonical_name = canonicalize_name(package_name)
        format_control = FormatControl(
            no_binary=canonical_name in self.no_binary or ":all:" in self.no_binary,
            only_binary=canonical_name in self.only_binary
            or ":all:" in self.only_binary,
        )
        return Evaluator(
            package_name=package_name,
            session=self.session,
            target_python=self.target_python,
            ignore_compatibility=self.ignore_compatibility,
            allow_yanked=allow_yanked,
            hashes=hashes or {},
            format_control=format_control,
        )

    def _build_index_page_link(self, index_url: str, package_name: str) -> Link:
        return Link(
            urljoin(index_url.rstrip("/") + "/", canonicalize_name(package_name) + "/")
        )

    def _build_find_link(self, find_link: str) -> Link:
        if os.path.exists(find_link):
            return Link.from_path(os.path.abspath(find_link))
        return Link(find_link)

    def _evaluate_links(
        self, links: Iterable[Link], evaluator: Evaluator
    ) -> Iterable[Package]:
        return filter(None, map(evaluator.evaluate_link, links))

    def _evaluate_packages(
        self,
        packages: Iterable[Package],
        requirement: packaging.requirements.Requirement,
        allow_prereleases: bool | None = None,
    ) -> Iterable[Package]:
        evaluator = functools.partial(
            evaluate_package,
            requirement=requirement,
            allow_prereleases=allow_prereleases,
        )
        return filter(evaluator, packages)

    def _sort_key(self, package: Package) -> tuple:
        """The key for sort, package with the largest value is the most preferred."""
        link = package.link
        pri = len(self._tag_priorities)
        build_tag: BuildTag = ()
        prefer_binary = False
        if link.is_wheel:
            *_, build_tag, file_tags = parse_wheel_filename(link.filename)
            pri = min(
                (self._tag_priorities.get(tag, pri) for tag in file_tags), default=pri
            )
            if self.prefer_binary:
                prefer_binary = True
        comes_from = package.link.comes_from
        source_index = len(self._source_order)

        if comes_from is not None and self.respect_source_order:
            source_index = next(
                (
                    i
                    for i, url in enumerate(self._source_order)
                    if comes_from.startswith(url)
                ),
                source_index,
            )
        return (
            -int(link.is_yanked),
            int(prefer_binary),
            -source_index,
            parse_version(package.version) if package.version is not None else 0,
            -pri,
            build_tag,
        )

    def _find_packages(
        self,
        package_name: str,
        allow_yanked: bool = False,
        hashes: dict[str, list[str]] | None = None,
    ) -> Iterable[Package]:
        """Find all packages with the given name.

        Args:
            package_name (str): The desired package name
            allow_yanked (bool): Whether to allow yanked candidates.
            hashes (dict[str, list[str]]|None): The hashes to filter on.

        Returns:
            Iterable[Package]: The packages with the given name
        """
        evaluator = self.build_evaluator(package_name, allow_yanked, hashes)
        for index_url in self.index_urls:
            package_link = self._build_index_page_link(index_url, package_name)
            yield from self._evaluate_links(
                collect_links_from_location(self.session, package_link), evaluator
            )
        for find_link in self.find_links:
            link = self._build_find_link(find_link)
            yield from self._evaluate_links(
                collect_links_from_location(self.session, link, expand=True), evaluator
            )

    def find_all_packages(
        self,
        package_name: str,
        allow_yanked: bool = False,
        hashes: dict[str, list[str]] | None = None,
    ) -> list[Package]:
        """Find all packages with the given package name, best match first.

        Args:
            package_name (str): The desired package name
            allow_yanked (bool): Whether to allow yanked candidates.
            hashes (dict[str, list[str]]|None): The hashes to filter on.

        Returns:
            list[Package]: The packages list sorted by best match
        """
        return sorted(
            self._find_packages(package_name, allow_yanked, hashes),
            key=self._sort_key,
            reverse=True,
        )

    def _find_packages_from_requirement(
        self,
        requirement: packaging.requirements.Requirement,
        allow_yanked: bool | None = None,
        hashes: dict[str, list[str]] | None = None,
    ) -> Iterable[Package]:
        if allow_yanked is None:
            allow_yanked = is_equality_specifier(requirement.specifier)
        if requirement.url:
            yield Package(requirement.name, None, link=Link(requirement.url))
        else:
            yield from self._find_packages(requirement.name, allow_yanked, hashes)

    def find_matches(
        self,
        requirement: packaging.requirements.Requirement | str,
        allow_yanked: bool | None = None,
        allow_prereleases: bool | None = None,
        hashes: dict[str, list[str]] | None = None,
    ) -> list[Package]:
        """Find all packages matching the given requirement, best match first.

        Args:
            requirement: A packaging.requirements.Requirement
                instance or a string to construct it.
            allow_yanked (bool|None): Whether to allow yanked candidates,
                or None to infer from the specifier.
            allow_prereleases (bool|None): Whether to allow prereleases,
                or None to infer from the specifier.
            hashes (dict[str, list[str]]|None): The hashes to filter on.

        Returns:
            list[Package]: The packages list sorted by best match
        """
        if isinstance(requirement, str):
            requirement = packaging.requirements.Requirement(requirement)
        return sorted(
            self._evaluate_packages(
                self._find_packages_from_requirement(requirement, allow_yanked, hashes),
                requirement,
                allow_prereleases,
            ),
            key=self._sort_key,
            reverse=True,
        )

    def find_best_match(
        self,
        requirement: packaging.requirements.Requirement | str,
        allow_yanked: bool | None = None,
        allow_prereleases: bool | None = None,
        hashes: dict[str, list[str]] | None = None,
    ) -> BestMatch:
        """Find the best match for the given requirement.

        Args:
            requirement: A packaging.requirements.Requirement
                instance or a string to construct it.
            allow_yanked (bool|None): Whether to allow yanked candidates,
                or None to infer from the specifier.
            allow_prereleases (bool|None): Whether to allow prereleases,
                or None to infer from the specifier.
            hashes (dict[str, list[str]]|None): The hashes to filter on.

        Returns:
            BestMatch: The best match
        """
        if isinstance(requirement, str):
            requirement = packaging.requirements.Requirement(requirement)
        candidates = list(
            self._find_packages_from_requirement(requirement, allow_yanked, hashes)
        )
        applicable_candidates = list(
            self._evaluate_packages(candidates, requirement, allow_prereleases)
        )
        best_match = max(applicable_candidates, key=self._sort_key, default=None)
        return BestMatch(best_match, applicable_candidates, candidates)

    def download_and_unpack(
        self,
        link: Link,
        location: str | pathlib.Path,
        download_dir: str | pathlib.Path | None = None,
        hashes: dict[str, list[str]] | None = None,
    ) -> pathlib.Path:
        """Download and unpack the package at the given link.

        If the link is a remote link, it will be downloaded to the ``download_dir``.
        Then, if the link is a wheel, the path of wheel file will be returned,
        otherwise it will be unpacked to the destination. Specially, if the link refers
        to a local directory, the path will be returned directly, otherwise the unpacked
        source path will be returned. And if the link has a subdirectory fragment, the
        subdirectory will be returned.

        Args:
            link (Link): The link to download
            location: The destination directory
            download_dir: The directory to download to, or None to use a
                temporary directory created by unearth.
            hashes (dict[str, list[str]]|None): The optional hash dict for validation.

        Returns:
            The path to the installable file or directory.
        """
        # Strip the rev part for VCS links
        if hashes is None and link.hash_name:
            hashes = {link.hash_name: [cast(str, link.hash)]}
        if download_dir is None:
            download_dir = TemporaryDirectory(prefix="unearth-download-").name
        file = unpack_link(
            self.session,
            link,
            pathlib.Path(download_dir),
            pathlib.Path(location),
            hashes,
            verbosity=self.verbosity,
        )
        return file.joinpath(link.subdirectory) if link.subdirectory else file
